@@ -13,18 +13,17 @@
    | license@zend.com so we can mail you a copy immediately.              |
    +----------------------------------------------------------------------+
    | Authors: Bob Weinand <bwoebi@php.net>                                |
-   |          Dmitry Stogov <dmitry@zend.com>                             |
+   |          Dmitry Stogov <dmitry@php.net>                              |
    +----------------------------------------------------------------------+
 */
 
-/* $Id$ */
-
 #include "zend_ast.h"
 #include "zend_API.h"
 #include "zend_operators.h"
 #include "zend_language_parser.h"
 #include "zend_smart_str.h"
 #include "zend_exceptions.h"
+#include "zend_constants.h"
 
 ZEND_API zend_ast_process_t zend_ast_process = NULL;
 
@@ -46,7 +45,7 @@ static inline size_t zend_ast_list_size(uint32_t children) {
 	return sizeof(zend_ast_list) - sizeof(zend_ast *) + sizeof(zend_ast *) * children;
 }
 
-ZEND_API zend_ast *zend_ast_create_znode(znode *node) {
+ZEND_API zend_ast * ZEND_FASTCALL zend_ast_create_znode(znode *node) {
 	zend_ast_znode *ast;
 
 	ast = zend_ast_alloc(sizeof(zend_ast_znode));
@@ -57,19 +56,50 @@ ZEND_API zend_ast *zend_ast_create_znode(znode *node) {
 	return (zend_ast *) ast;
 }
 
-ZEND_API zend_ast *zend_ast_create_zval_with_lineno(zval *zv, zend_ast_attr attr, uint32_t lineno) {
+static zend_always_inline zend_ast * zend_ast_create_zval_int(zval *zv, uint32_t attr, uint32_t lineno) {
 	zend_ast_zval *ast;
 
 	ast = zend_ast_alloc(sizeof(zend_ast_zval));
 	ast->kind = ZEND_AST_ZVAL;
 	ast->attr = attr;
 	ZVAL_COPY_VALUE(&ast->val, zv);
-	ast->val.u2.lineno = lineno;
+	Z_LINENO(ast->val) = lineno;
 	return (zend_ast *) ast;
 }
 
-ZEND_API zend_ast *zend_ast_create_zval_ex(zval *zv, zend_ast_attr attr) {
-	return zend_ast_create_zval_with_lineno(zv, attr, CG(zend_lineno));
+ZEND_API zend_ast * ZEND_FASTCALL zend_ast_create_zval_with_lineno(zval *zv, uint32_t lineno) {
+	return zend_ast_create_zval_int(zv, 0, lineno);
+}
+
+ZEND_API zend_ast * ZEND_FASTCALL zend_ast_create_zval_ex(zval *zv, zend_ast_attr attr) {
+	return zend_ast_create_zval_int(zv, attr, CG(zend_lineno));
+}
+
+ZEND_API zend_ast * ZEND_FASTCALL zend_ast_create_zval(zval *zv) {
+	return zend_ast_create_zval_int(zv, 0, CG(zend_lineno));
+}
+
+ZEND_API zend_ast * ZEND_FASTCALL zend_ast_create_zval_from_str(zend_string *str) {
+	zval zv;
+	ZVAL_STR(&zv, str);
+	return zend_ast_create_zval_int(&zv, 0, CG(zend_lineno));
+}
+
+ZEND_API zend_ast * ZEND_FASTCALL zend_ast_create_zval_from_long(zend_long lval) {
+	zval zv;
+	ZVAL_LONG(&zv, lval);
+	return zend_ast_create_zval_int(&zv, 0, CG(zend_lineno));
+}
+
+ZEND_API zend_ast * ZEND_FASTCALL zend_ast_create_constant(zend_string *name, zend_ast_attr attr) {
+	zend_ast_zval *ast;
+
+	ast = zend_ast_alloc(sizeof(zend_ast_zval));
+	ast->kind = ZEND_AST_CONSTANT;
+	ast->attr = attr;
+	ZVAL_STR(&ast->val, name);
+	Z_LINENO(ast->val) = CG(zend_lineno);
+	return (zend_ast *) ast;
 }
 
 ZEND_API zend_ast *zend_ast_create_decl(
@@ -95,6 +125,183 @@ ZEND_API zend_ast *zend_ast_create_decl(
 	return (zend_ast *) ast;
 }
 
+#if ZEND_AST_SPEC
+ZEND_API zend_ast * ZEND_FASTCALL zend_ast_create_0(zend_ast_kind kind) {
+	zend_ast *ast;
+
+	ZEND_ASSERT(kind >> ZEND_AST_NUM_CHILDREN_SHIFT == 0);
+	ast = zend_ast_alloc(zend_ast_size(0));
+	ast->kind = kind;
+	ast->attr = 0;
+	ast->lineno = CG(zend_lineno);
+
+	return ast;
+}
+
+ZEND_API zend_ast * ZEND_FASTCALL zend_ast_create_1(zend_ast_kind kind, zend_ast *child) {
+	zend_ast *ast;
+	uint32_t lineno;
+
+	ZEND_ASSERT(kind >> ZEND_AST_NUM_CHILDREN_SHIFT == 1);
+	ast = zend_ast_alloc(zend_ast_size(1));
+	ast->kind = kind;
+	ast->attr = 0;
+	ast->child[0] = child;
+	if (child) {
+		lineno = zend_ast_get_lineno(child);
+	} else {
+		lineno = CG(zend_lineno);
+	}
+	ast->lineno = lineno;
+	ast->lineno = lineno;
+
+	return ast;
+}
+
+ZEND_API zend_ast * ZEND_FASTCALL zend_ast_create_2(zend_ast_kind kind, zend_ast *child1, zend_ast *child2) {
+	zend_ast *ast;
+	uint32_t lineno;
+
+	ZEND_ASSERT(kind >> ZEND_AST_NUM_CHILDREN_SHIFT == 2);
+	ast = zend_ast_alloc(zend_ast_size(2));
+	ast->kind = kind;
+	ast->attr = 0;
+	ast->child[0] = child1;
+	ast->child[1] = child2;
+	if (child1) {
+		lineno = zend_ast_get_lineno(child1);
+	} else if (child2) {
+		lineno = zend_ast_get_lineno(child2);
+	} else {
+		lineno = CG(zend_lineno);
+	}
+	ast->lineno = lineno;
+
+	return ast;
+}
+
+ZEND_API zend_ast * ZEND_FASTCALL zend_ast_create_3(zend_ast_kind kind, zend_ast *child1, zend_ast *child2, zend_ast *child3) {
+	zend_ast *ast;
+	uint32_t lineno;
+
+	ZEND_ASSERT(kind >> ZEND_AST_NUM_CHILDREN_SHIFT == 3);
+	ast = zend_ast_alloc(zend_ast_size(3));
+	ast->kind = kind;
+	ast->attr = 0;
+	ast->child[0] = child1;
+	ast->child[1] = child2;
+	ast->child[2] = child3;
+	if (child1) {
+		lineno = zend_ast_get_lineno(child1);
+	} else if (child2) {
+		lineno = zend_ast_get_lineno(child2);
+	} else if (child3) {
+		lineno = zend_ast_get_lineno(child3);
+	} else {
+		lineno = CG(zend_lineno);
+	}
+	ast->lineno = lineno;
+
+	return ast;
+}
+
+ZEND_API zend_ast * ZEND_FASTCALL zend_ast_create_4(zend_ast_kind kind, zend_ast *child1, zend_ast *child2, zend_ast *child3, zend_ast *child4) {
+	zend_ast *ast;
+	uint32_t lineno;
+
+	ZEND_ASSERT(kind >> ZEND_AST_NUM_CHILDREN_SHIFT == 4);
+	ast = zend_ast_alloc(zend_ast_size(4));
+	ast->kind = kind;
+	ast->attr = 0;
+	ast->child[0] = child1;
+	ast->child[1] = child2;
+	ast->child[2] = child3;
+	ast->child[3] = child4;
+	if (child1) {
+		lineno = zend_ast_get_lineno(child1);
+	} else if (child2) {
+		lineno = zend_ast_get_lineno(child2);
+	} else if (child3) {
+		lineno = zend_ast_get_lineno(child3);
+	} else if (child4) {
+		lineno = zend_ast_get_lineno(child4);
+	} else {
+		lineno = CG(zend_lineno);
+	}
+	ast->lineno = lineno;
+
+	return ast;
+}
+
+ZEND_API zend_ast * ZEND_FASTCALL zend_ast_create_list_0(zend_ast_kind kind) {
+	zend_ast *ast;
+	zend_ast_list *list;
+
+	ast = zend_ast_alloc(zend_ast_list_size(4));
+	list = (zend_ast_list *) ast;
+	list->kind = kind;
+	list->attr = 0;
+	list->lineno = CG(zend_lineno);
+	list->children = 0;
+
+	return ast;
+}
+
+ZEND_API zend_ast * ZEND_FASTCALL zend_ast_create_list_1(zend_ast_kind kind, zend_ast *child) {
+	zend_ast *ast;
+	zend_ast_list *list;
+	uint32_t lineno;
+
+	ast = zend_ast_alloc(zend_ast_list_size(4));
+	list = (zend_ast_list *) ast;
+	list->kind = kind;
+	list->attr = 0;
+	list->children = 1;
+	list->child[0] = child;
+	if (child) {
+		lineno = zend_ast_get_lineno(child);
+		if (lineno > CG(zend_lineno)) {
+			lineno = CG(zend_lineno);
+		}
+	} else {
+		lineno = CG(zend_lineno);
+	}
+	list->lineno = lineno;
+
+	return ast;
+}
+
+ZEND_API zend_ast * ZEND_FASTCALL zend_ast_create_list_2(zend_ast_kind kind, zend_ast *child1, zend_ast *child2) {
+	zend_ast *ast;
+	zend_ast_list *list;
+	uint32_t lineno;
+
+	ast = zend_ast_alloc(zend_ast_list_size(4));
+	list = (zend_ast_list *) ast;
+	list->kind = kind;
+	list->attr = 0;
+	list->children = 2;
+	list->child[0] = child1;
+	list->child[1] = child2;
+	if (child1) {
+		lineno = zend_ast_get_lineno(child1);
+		if (lineno > CG(zend_lineno)) {
+			lineno = CG(zend_lineno);
+		}
+	} else if (child2) {
+		lineno = zend_ast_get_lineno(child2);
+		if (lineno > CG(zend_lineno)) {
+			lineno = CG(zend_lineno);
+		}
+	} else {
+		list->children = 0;
+		lineno = CG(zend_lineno);
+	}
+	list->lineno = lineno;
+
+	return ast;
+}
+#else
 static zend_ast *zend_ast_create_from_va_list(zend_ast_kind kind, zend_ast_attr attr, va_list va) {
 	uint32_t i, children = kind >> ZEND_AST_NUM_CHILDREN_SHIFT;
 	zend_ast *ast;
@@ -173,12 +380,13 @@ ZEND_API zend_ast *zend_ast_create_list(uint32_t init_children, zend_ast_kind ki
 
 	return ast;
 }
+#endif
 
 static inline zend_bool is_power_of_two(uint32_t n) {
 	return ((n != 0) && (n == (n & (~n + 1))));
 }
 
-ZEND_API zend_ast *zend_ast_list_add(zend_ast *ast, zend_ast *op) {
+ZEND_API zend_ast * ZEND_FASTCALL zend_ast_list_add(zend_ast *ast, zend_ast *op) {
 	zend_ast_list *list = zend_ast_get_list(ast);
 	if (list->children >= 4 && is_power_of_two(list->children)) {
 			list = zend_ast_realloc(list,
@@ -195,12 +403,12 @@ static int zend_ast_add_array_element(zval *result, zval *offset, zval *expr)
 			if (!zend_hash_next_index_insert(Z_ARRVAL_P(result), expr)) {
 				zend_error(E_WARNING,
 					"Cannot add element to the array as the next element is already occupied");
-				zval_ptr_dtor(expr);
+				zval_ptr_dtor_nogc(expr);
 			}
 			break;
 		case IS_STRING:
 			zend_symtable_update(Z_ARRVAL_P(result), Z_STR_P(offset), expr);
-			zval_dtor(offset);
+			zval_ptr_dtor_str(offset);
 			break;
 		case IS_NULL:
 			zend_symtable_update(Z_ARRVAL_P(result), ZSTR_EMPTY_ALLOC(), expr);
@@ -228,7 +436,7 @@ static int zend_ast_add_array_element(zval *result, zval *offset, zval *expr)
 	return SUCCESS;
 }
 
-ZEND_API int zend_ast_evaluate(zval *result, zend_ast *ast, zend_class_entry *scope)
+ZEND_API int ZEND_FASTCALL zend_ast_evaluate(zval *result, zend_ast *ast, zend_class_entry *scope)
 {
 	zval op1, op2;
 	int ret = SUCCESS;
@@ -238,13 +446,13 @@ ZEND_API int zend_ast_evaluate(zval *result, zend_ast *ast, zend_class_entry *sc
 			if (UNEXPECTED(zend_ast_evaluate(&op1, ast->child[0], scope) != SUCCESS)) {
 				ret = FAILURE;
 			} else if (UNEXPECTED(zend_ast_evaluate(&op2, ast->child[1], scope) != SUCCESS)) {
-				zval_dtor(&op1);
+				zval_ptr_dtor_nogc(&op1);
 				ret = FAILURE;
 			} else {
 				binary_op_type op = get_binary_op(ast->attr);
 				ret = op(result, &op1, &op2);
-				zval_dtor(&op1);
-				zval_dtor(&op2);
+				zval_ptr_dtor_nogc(&op1);
+				zval_ptr_dtor_nogc(&op2);
 			}
 			break;
 		case ZEND_AST_GREATER:
@@ -252,15 +460,15 @@ ZEND_API int zend_ast_evaluate(zval *result, zend_ast *ast, zend_class_entry *sc
 			if (UNEXPECTED(zend_ast_evaluate(&op1, ast->child[0], scope) != SUCCESS)) {
 				ret = FAILURE;
 			} else if (UNEXPECTED(zend_ast_evaluate(&op2, ast->child[1], scope) != SUCCESS)) {
-				zval_dtor(&op1);
+				zval_ptr_dtor_nogc(&op1);
 				ret = FAILURE;
 			} else {
 				/* op1 > op2 is the same as op2 < op1 */
 				binary_op_type op = ast->kind == ZEND_AST_GREATER
 					? is_smaller_function : is_smaller_or_equal_function;
 				ret = op(result, &op2, &op1);
-				zval_dtor(&op1);
-				zval_dtor(&op2);
+				zval_ptr_dtor_nogc(&op1);
+				zval_ptr_dtor_nogc(&op2);
 			}
 			break;
 		case ZEND_AST_UNARY_OP:
@@ -269,32 +477,37 @@ ZEND_API int zend_ast_evaluate(zval *result, zend_ast *ast, zend_class_entry *sc
 			} else {
 				unary_op_type op = get_unary_op(ast->attr);
 				ret = op(result, &op1);
-				zval_dtor(&op1);
+				zval_ptr_dtor_nogc(&op1);
 			}
 			break;
 		case ZEND_AST_ZVAL:
 		{
 			zval *zv = zend_ast_get_zval(ast);
 
-			if (Z_OPT_CONSTANT_P(zv)) {
-				if (Z_TYPE_FLAGS_P(zv) & IS_TYPE_REFCOUNTED) {
-					if (UNEXPECTED(zval_update_constant_ex(zv, scope) != SUCCESS)) {
-						ret = FAILURE;
-						break;
-					}
-					ZVAL_COPY(result, zv);
-				} else {
-					ZVAL_COPY_VALUE(result, zv);
-					if (UNEXPECTED(zval_update_constant_ex(result, scope) != SUCCESS)) {
-						ret = FAILURE;
-						break;
-					}
-				}
-			} else {
-				ZVAL_COPY(result, zv);
-			}
+			ZVAL_COPY(result, zv);
 			break;
 		}
+		case ZEND_AST_CONSTANT:
+		{
+			zend_string *name = zend_ast_get_constant_name(ast);
+			zval *zv = zend_get_constant_ex(name, scope, ast->attr);
+
+			if (UNEXPECTED(zv == NULL)) {
+				ZVAL_UNDEF(result);
+				ret = zend_use_undefined_constant(name, ast->attr, result);
+				break;
+			}
+			ZVAL_COPY_OR_DUP(result, zv);
+			break;
+		}
+		case ZEND_AST_CONSTANT_CLASS:
+			ZEND_ASSERT(EG(current_execute_data));
+			if (scope && scope->name) {
+				ZVAL_STR_COPY(result, scope->name);
+			} else {
+				ZVAL_EMPTY_STRING(result);
+			}
+			break;
 		case ZEND_AST_AND:
 			if (UNEXPECTED(zend_ast_evaluate(&op1, ast->child[0], scope) != SUCCESS)) {
 				ret = FAILURE;
@@ -302,16 +515,16 @@ ZEND_API int zend_ast_evaluate(zval *result, zend_ast *ast, zend_class_entry *sc
 			}
 			if (zend_is_true(&op1)) {
 				if (UNEXPECTED(zend_ast_evaluate(&op2, ast->child[1], scope) != SUCCESS)) {
-					zval_dtor(&op1);
+					zval_ptr_dtor_nogc(&op1);
 					ret = FAILURE;
 					break;
 				}
 				ZVAL_BOOL(result, zend_is_true(&op2));
-				zval_dtor(&op2);
+				zval_ptr_dtor_nogc(&op2);
 			} else {
 				ZVAL_FALSE(result);
 			}
-			zval_dtor(&op1);
+			zval_ptr_dtor_nogc(&op1);
 			break;
 		case ZEND_AST_OR:
 			if (UNEXPECTED(zend_ast_evaluate(&op1, ast->child[0], scope) != SUCCESS)) {
@@ -322,14 +535,14 @@ ZEND_API int zend_ast_evaluate(zval *result, zend_ast *ast, zend_class_entry *sc
 				ZVAL_TRUE(result);
 			} else {
 				if (UNEXPECTED(zend_ast_evaluate(&op2, ast->child[1], scope) != SUCCESS)) {
-					zval_dtor(&op1);
+					zval_ptr_dtor_nogc(&op1);
 					ret = FAILURE;
 					break;
 				}
 				ZVAL_BOOL(result, zend_is_true(&op2));
-				zval_dtor(&op2);
+				zval_ptr_dtor_nogc(&op2);
 			}
-			zval_dtor(&op1);
+			zval_ptr_dtor_nogc(&op1);
 			break;
 		case ZEND_AST_CONDITIONAL:
 			if (UNEXPECTED(zend_ast_evaluate(&op1, ast->child[0], scope) != SUCCESS)) {
@@ -341,19 +554,19 @@ ZEND_API int zend_ast_evaluate(zval *result, zend_ast *ast, zend_class_entry *sc
 					*result = op1;
 				} else {
 					if (UNEXPECTED(zend_ast_evaluate(result, ast->child[1], scope) != SUCCESS)) {
-						zval_dtor(&op1);
+						zval_ptr_dtor_nogc(&op1);
 						ret = FAILURE;
 						break;
 					}
-					zval_dtor(&op1);
+					zval_ptr_dtor_nogc(&op1);
 				}
 			} else {
 				if (UNEXPECTED(zend_ast_evaluate(result, ast->child[2], scope) != SUCCESS)) {
-					zval_dtor(&op1);
+					zval_ptr_dtor_nogc(&op1);
 					ret = FAILURE;
 					break;
 				}
-				zval_dtor(&op1);
+				zval_ptr_dtor_nogc(&op1);
 			}
 			break;
 		case ZEND_AST_COALESCE:
@@ -365,11 +578,11 @@ ZEND_API int zend_ast_evaluate(zval *result, zend_ast *ast, zend_class_entry *sc
 				*result = op1;
 			} else {
 				if (UNEXPECTED(zend_ast_evaluate(result, ast->child[1], scope) != SUCCESS)) {
-					zval_dtor(&op1);
+					zval_ptr_dtor_nogc(&op1);
 					ret = FAILURE;
 					break;
 				}
-				zval_dtor(&op1);
+				zval_ptr_dtor_nogc(&op1);
 			}
 			break;
 		case ZEND_AST_UNARY_PLUS:
@@ -378,7 +591,7 @@ ZEND_API int zend_ast_evaluate(zval *result, zend_ast *ast, zend_class_entry *sc
 			} else {
 				ZVAL_LONG(&op1, 0);
 				ret = add_function(result, &op1, &op2);
-				zval_dtor(&op2);
+				zval_ptr_dtor_nogc(&op2);
 			}
 			break;
 		case ZEND_AST_UNARY_MINUS:
@@ -387,33 +600,38 @@ ZEND_API int zend_ast_evaluate(zval *result, zend_ast *ast, zend_class_entry *sc
 			} else {
 				ZVAL_LONG(&op1, 0);
 				ret = sub_function(result, &op1, &op2);
-				zval_dtor(&op2);
+				zval_ptr_dtor_nogc(&op2);
 			}
 			break;
 		case ZEND_AST_ARRAY:
-			array_init(result);
 			{
 				uint32_t i;
 				zend_ast_list *list = zend_ast_get_list(ast);
+
+				if (!list->children) {
+					ZVAL_EMPTY_ARRAY(result);
+					break;
+				}
+				array_init(result);
 				for (i = 0; i < list->children; i++) {
 					zend_ast *elem = list->child[i];
 					if (elem->child[1]) {
 						if (UNEXPECTED(zend_ast_evaluate(&op1, elem->child[1], scope) != SUCCESS)) {
-							zval_dtor(result);
+							zval_ptr_dtor_nogc(result);
 							return FAILURE;
 						}
 					} else {
 						ZVAL_UNDEF(&op1);
 					}
 					if (UNEXPECTED(zend_ast_evaluate(&op2, elem->child[0], scope) != SUCCESS)) {
-						zval_dtor(&op1);
-						zval_dtor(result);
+						zval_ptr_dtor_nogc(&op1);
+						zval_ptr_dtor_nogc(result);
 						return FAILURE;
 					}
 					if (UNEXPECTED(zend_ast_add_array_element(result, &op1, &op2) != SUCCESS)) {
-						zval_dtor(&op1);
-						zval_dtor(&op2);
-						zval_dtor(result);
+						zval_ptr_dtor_nogc(&op1);
+						zval_ptr_dtor_nogc(&op2);
+						zval_ptr_dtor_nogc(result);
 						return FAILURE;
 					}
 				}
@@ -427,21 +645,13 @@ ZEND_API int zend_ast_evaluate(zval *result, zend_ast *ast, zend_class_entry *sc
 			if (UNEXPECTED(zend_ast_evaluate(&op1, ast->child[0], scope) != SUCCESS)) {
 				ret = FAILURE;
 			} else if (UNEXPECTED(zend_ast_evaluate(&op2, ast->child[1], scope) != SUCCESS)) {
-				zval_dtor(&op1);
+				zval_ptr_dtor_nogc(&op1);
 				ret = FAILURE;
 			} else {
-				zval tmp;
+				zend_fetch_dimension_const(result, &op1, &op2, (ast->attr == ZEND_DIM_IS) ? BP_VAR_IS : BP_VAR_R);
 
-				zend_fetch_dimension_const(&tmp, &op1, &op2, (ast->attr == ZEND_DIM_IS) ? BP_VAR_IS : BP_VAR_R);
-
-				if (UNEXPECTED(Z_ISREF(tmp))) {
-					ZVAL_DUP(result, Z_REFVAL(tmp));
-				} else {
-					ZVAL_DUP(result, &tmp);
-				}
-				zval_ptr_dtor(&tmp);
-				zval_dtor(&op1);
-				zval_dtor(&op2);
+				zval_ptr_dtor_nogc(&op1);
+				zval_ptr_dtor_nogc(&op2);
 			}
 			break;
 		default:
@@ -451,94 +661,148 @@ ZEND_API int zend_ast_evaluate(zval *result, zend_ast *ast, zend_class_entry *sc
 	return ret;
 }
 
-ZEND_API zend_ast *zend_ast_copy(zend_ast *ast)
+static size_t ZEND_FASTCALL zend_ast_tree_size(zend_ast *ast)
 {
-	if (ast == NULL) {
-		return NULL;
-	} else if (ast->kind == ZEND_AST_ZVAL) {
-		zend_ast_zval *new = emalloc(sizeof(zend_ast_zval));
+	size_t size;
+
+	if (ast->kind == ZEND_AST_ZVAL || ast->kind == ZEND_AST_CONSTANT) {
+		size = sizeof(zend_ast_zval);
+	} else if (zend_ast_is_list(ast)) {
+		uint32_t i;
+		zend_ast_list *list = zend_ast_get_list(ast);
+
+		size = zend_ast_list_size(list->children);
+		for (i = 0; i < list->children; i++) {
+			if (list->child[i]) {
+				size += zend_ast_tree_size(list->child[i]);
+			}
+		}
+	} else {
+		uint32_t i, children = zend_ast_get_num_children(ast);
+
+		size = zend_ast_size(children);
+		for (i = 0; i < children; i++) {
+			if (ast->child[i]) {
+				size += zend_ast_tree_size(ast->child[i]);
+			}
+		}
+	}
+	return size;
+}
+
+static void* ZEND_FASTCALL zend_ast_tree_copy(zend_ast *ast, void *buf)
+{
+	if (ast->kind == ZEND_AST_ZVAL) {
+		zend_ast_zval *new = (zend_ast_zval*)buf;
 		new->kind = ZEND_AST_ZVAL;
 		new->attr = ast->attr;
 		ZVAL_COPY(&new->val, zend_ast_get_zval(ast));
-		return (zend_ast *) new;
+		buf = (void*)((char*)buf + sizeof(zend_ast_zval));
+	} else if (ast->kind == ZEND_AST_CONSTANT) {
+		zend_ast_zval *new = (zend_ast_zval*)buf;
+		new->kind = ZEND_AST_CONSTANT;
+		new->attr = ast->attr;
+		ZVAL_STR_COPY(&new->val, zend_ast_get_constant_name(ast));
+		buf = (void*)((char*)buf + sizeof(zend_ast_zval));
 	} else if (zend_ast_is_list(ast)) {
 		zend_ast_list *list = zend_ast_get_list(ast);
-		zend_ast_list *new = emalloc(zend_ast_list_size(list->children));
+		zend_ast_list *new = (zend_ast_list*)buf;
 		uint32_t i;
 		new->kind = list->kind;
 		new->attr = list->attr;
 		new->children = list->children;
+		buf = (void*)((char*)buf + zend_ast_list_size(list->children));
 		for (i = 0; i < list->children; i++) {
-			new->child[i] = zend_ast_copy(list->child[i]);
+			if (list->child[i]) {
+				new->child[i] = (zend_ast*)buf;
+				buf = zend_ast_tree_copy(list->child[i], buf);
+			} else {
+				new->child[i] = NULL;
+			}
 		}
-		return (zend_ast *) new;
 	} else {
 		uint32_t i, children = zend_ast_get_num_children(ast);
-		zend_ast *new = emalloc(zend_ast_size(children));
+		zend_ast *new = (zend_ast*)buf;
 		new->kind = ast->kind;
 		new->attr = ast->attr;
+		buf = (void*)((char*)buf + zend_ast_size(children));
 		for (i = 0; i < children; i++) {
-			new->child[i] = zend_ast_copy(ast->child[i]);
+			if (ast->child[i]) {
+				new->child[i] = (zend_ast*)buf;
+				buf = zend_ast_tree_copy(ast->child[i], buf);
+			} else {
+				new->child[i] = NULL;
+			}
 		}
-		return new;
 	}
+	return buf;
 }
 
-static void zend_ast_destroy_ex(zend_ast *ast, zend_bool free) {
+ZEND_API zend_ast_ref * ZEND_FASTCALL zend_ast_copy(zend_ast *ast)
+{
+	size_t tree_size;
+	zend_ast_ref *ref;
+
+	ZEND_ASSERT(ast != NULL);
+	tree_size = zend_ast_tree_size(ast) + sizeof(zend_ast_ref);
+	ref = emalloc(tree_size);
+	zend_ast_tree_copy(ast, GC_AST(ref));
+	GC_SET_REFCOUNT(ref, 1);
+	GC_TYPE_INFO(ref) = IS_CONSTANT_AST;
+	return ref;
+}
+
+ZEND_API void ZEND_FASTCALL zend_ast_destroy(zend_ast *ast)
+{
+tail_call:
 	if (!ast) {
 		return;
 	}
 
-	switch (ast->kind) {
-		case ZEND_AST_ZVAL:
-			/* Destroy value without using GC: When opcache moves arrays into SHM it will
-			 * free the zend_array structure, so references to it from outside the op array
-			 * become invalid. GC would cause such a reference in the root buffer. */
-			zval_ptr_dtor_nogc(zend_ast_get_zval(ast));
-			break;
-		case ZEND_AST_FUNC_DECL:
-		case ZEND_AST_CLOSURE:
-		case ZEND_AST_METHOD:
-		case ZEND_AST_CLASS:
-		{
-			zend_ast_decl *decl = (zend_ast_decl *) ast;
-			if (decl->name) {
-			    zend_string_release(decl->name);
-			}
-			if (decl->doc_comment) {
-				zend_string_release(decl->doc_comment);
-			}
-			zend_ast_destroy_ex(decl->child[0], free);
-			zend_ast_destroy_ex(decl->child[1], free);
-			zend_ast_destroy_ex(decl->child[2], free);
-			zend_ast_destroy_ex(decl->child[3], free);
-			break;
+	if (EXPECTED(ast->kind >= ZEND_AST_VAR)) {
+		uint32_t i, children = zend_ast_get_num_children(ast);
+
+		for (i = 1; i < children; i++) {
+			zend_ast_destroy(ast->child[i]);
 		}
-		default:
-			if (zend_ast_is_list(ast)) {
-				zend_ast_list *list = zend_ast_get_list(ast);
-				uint32_t i;
-				for (i = 0; i < list->children; i++) {
-					zend_ast_destroy_ex(list->child[i], free);
-				}
-			} else {
-				uint32_t i, children = zend_ast_get_num_children(ast);
-				for (i = 0; i < children; i++) {
-					zend_ast_destroy_ex(ast->child[i], free);
-				}
+		ast = ast->child[0];
+		goto tail_call;
+	} else if (EXPECTED(ast->kind == ZEND_AST_ZVAL)) {
+		zval_ptr_dtor_nogc(zend_ast_get_zval(ast));
+	} else if (EXPECTED(zend_ast_is_list(ast))) {
+		zend_ast_list *list = zend_ast_get_list(ast);
+		if (list->children) {
+			uint32_t i;
+
+			for (i = 1; i < list->children; i++) {
+				zend_ast_destroy(list->child[i]);
 			}
-	}
+			ast = list->child[0];
+			goto tail_call;
+		}
+	} else if (EXPECTED(ast->kind == ZEND_AST_CONSTANT)) {
+		zend_string_release_ex(zend_ast_get_constant_name(ast), 0);
+	} else if (EXPECTED(ast->kind >= ZEND_AST_FUNC_DECL)) {
+		zend_ast_decl *decl = (zend_ast_decl *) ast;
 
-	if (free) {
-		efree(ast);
+		if (decl->name) {
+		    zend_string_release_ex(decl->name, 0);
+		}
+		if (decl->doc_comment) {
+			zend_string_release_ex(decl->doc_comment, 0);
+		}
+		zend_ast_destroy(decl->child[0]);
+		zend_ast_destroy(decl->child[1]);
+		zend_ast_destroy(decl->child[2]);
+		ast = decl->child[3];
+		goto tail_call;
 	}
 }
 
-ZEND_API void zend_ast_destroy(zend_ast *ast) {
-	zend_ast_destroy_ex(ast, 0);
-}
-ZEND_API void zend_ast_destroy_and_free(zend_ast *ast) {
-	zend_ast_destroy_ex(ast, 1);
+ZEND_API void ZEND_FASTCALL zend_ast_ref_destroy(zend_ast_ref *ast)
+{
+	zend_ast_destroy(GC_AST(ast));
+	efree(ast);
 }
 
 ZEND_API void zend_ast_apply(zend_ast *ast, zend_ast_apply_func fn) {
@@ -591,9 +855,9 @@ ZEND_API void zend_ast_apply(zend_ast *ast, zend_ast_apply_func fn) {
  *  270     non-associative clone new
  */
 
-static void zend_ast_export_ex(smart_str *str, zend_ast *ast, int priority, int indent);
+static ZEND_COLD void zend_ast_export_ex(smart_str *str, zend_ast *ast, int priority, int indent);
 
-static void zend_ast_export_str(smart_str *str, zend_string *s)
+static ZEND_COLD void zend_ast_export_str(smart_str *str, zend_string *s)
 {
 	size_t i;
 
@@ -608,7 +872,7 @@ static void zend_ast_export_str(smart_str *str, zend_string *s)
 	}
 }
 
-static void zend_ast_export_qstr(smart_str *str, char quote, zend_string *s)
+static ZEND_COLD void zend_ast_export_qstr(smart_str *str, char quote, zend_string *s)
 {
 	size_t i;
 
@@ -653,7 +917,7 @@ static void zend_ast_export_qstr(smart_str *str, char quote, zend_string *s)
 	}
 }
 
-static void zend_ast_export_indent(smart_str *str, int indent)
+static ZEND_COLD void zend_ast_export_indent(smart_str *str, int indent)
 {
 	while (indent > 0) {
 		smart_str_appends(str, "    ");
@@ -661,7 +925,7 @@ static void zend_ast_export_indent(smart_str *str, int indent)
 	}
 }
 
-static void zend_ast_export_name(smart_str *str, zend_ast *ast, int priority, int indent)
+static ZEND_COLD void zend_ast_export_name(smart_str *str, zend_ast *ast, int priority, int indent)
 {
 	if (ast->kind == ZEND_AST_ZVAL) {
 		zval *zv = zend_ast_get_zval(ast);
@@ -674,7 +938,7 @@ static void zend_ast_export_name(smart_str *str, zend_ast *ast, int priority, in
 	zend_ast_export_ex(str, ast, priority, indent);
 }
 
-static void zend_ast_export_ns_name(smart_str *str, zend_ast *ast, int priority, int indent)
+static ZEND_COLD void zend_ast_export_ns_name(smart_str *str, zend_ast *ast, int priority, int indent)
 {
 	if (ast->kind == ZEND_AST_ZVAL) {
 		zval *zv = zend_ast_get_zval(ast);
@@ -692,7 +956,7 @@ static void zend_ast_export_ns_name(smart_str *str, zend_ast *ast, int priority,
 	zend_ast_export_ex(str, ast, priority, indent);
 }
 
-static int zend_ast_valid_var_char(char ch)
+static ZEND_COLD int zend_ast_valid_var_char(char ch)
 {
 	unsigned char c = (unsigned char)ch;
 
@@ -705,7 +969,7 @@ static int zend_ast_valid_var_char(char ch)
 	return 1;
 }
 
-static int zend_ast_valid_var_name(const char *s, size_t len)
+static ZEND_COLD int zend_ast_valid_var_name(const char *s, size_t len)
 {
 	unsigned char c;
 	size_t i;
@@ -736,7 +1000,7 @@ static ZEND_COLD int zend_ast_var_needs_braces(char ch)
 	return ch == '[' || zend_ast_valid_var_char(ch);
 }
 
-static void zend_ast_export_var(smart_str *str, zend_ast *ast, int priority, int indent)
+static ZEND_COLD void zend_ast_export_var(smart_str *str, zend_ast *ast, int priority, int indent)
 {
 	if (ast->kind == ZEND_AST_ZVAL) {
 		zval *zv = zend_ast_get_zval(ast);
@@ -754,7 +1018,7 @@ static void zend_ast_export_var(smart_str *str, zend_ast *ast, int priority, int
 	smart_str_appendc(str, '}');
 }
 
-static void zend_ast_export_list(smart_str *str, zend_ast_list *list, int separator, int priority, int indent)
+static ZEND_COLD void zend_ast_export_list(smart_str *str, zend_ast_list *list, int separator, int priority, int indent)
 {
 	uint32_t i = 0;
 
@@ -767,7 +1031,7 @@ static void zend_ast_export_list(smart_str *str, zend_ast_list *list, int separa
 	}
 }
 
-static void zend_ast_export_encaps_list(smart_str *str, char quote, zend_ast_list *list, int indent)
+static ZEND_COLD void zend_ast_export_encaps_list(smart_str *str, char quote, zend_ast_list *list, int indent)
 {
 	uint32_t i = 0;
 	zend_ast *ast;
@@ -796,7 +1060,7 @@ static void zend_ast_export_encaps_list(smart_str *str, char quote, zend_ast_lis
 	}
 }
 
-static void zend_ast_export_name_list_ex(smart_str *str, zend_ast_list *list, int indent, const char *separator)
+static ZEND_COLD void zend_ast_export_name_list_ex(smart_str *str, zend_ast_list *list, int indent, const char *separator)
 {
 	uint32_t i = 0;
 
@@ -812,7 +1076,7 @@ static void zend_ast_export_name_list_ex(smart_str *str, zend_ast_list *list, in
 #define zend_ast_export_name_list(s, l, i) zend_ast_export_name_list_ex(s, l, i, ", ")
 #define zend_ast_export_catch_name_list(s, l, i) zend_ast_export_name_list_ex(s, l, i, "|")
 
-static void zend_ast_export_var_list(smart_str *str, zend_ast_list *list, int indent)
+static ZEND_COLD void zend_ast_export_var_list(smart_str *str, zend_ast_list *list, int indent)
 {
 	uint32_t i = 0;
 
@@ -829,7 +1093,7 @@ static void zend_ast_export_var_list(smart_str *str, zend_ast_list *list, int in
 	}
 }
 
-static void zend_ast_export_stmt(smart_str *str, zend_ast *ast, int indent)
+static ZEND_COLD void zend_ast_export_stmt(smart_str *str, zend_ast *ast, int indent)
 {
 	if (!ast) {
 		return;
@@ -871,7 +1135,7 @@ static void zend_ast_export_stmt(smart_str *str, zend_ast *ast, int indent)
 	}
 }
 
-static void zend_ast_export_if_stmt(smart_str *str, zend_ast_list *list, int indent)
+static ZEND_COLD void zend_ast_export_if_stmt(smart_str *str, zend_ast_list *list, int indent)
 {
 	uint32_t i;
 	zend_ast *ast;
@@ -894,7 +1158,7 @@ tail_call:
 		} else {
 			zend_ast_export_indent(str, indent);
 			smart_str_appends(str, "} else ");
-			if (ast->child[1]->kind == ZEND_AST_IF) {
+			if (ast->child[1] && ast->child[1]->kind == ZEND_AST_IF) {
 				list = (zend_ast_list*)ast->child[1];
 				goto tail_call;
 			} else {
@@ -908,7 +1172,7 @@ tail_call:
 	smart_str_appendc(str, '}');
 }
 
-static void zend_ast_export_zval(smart_str *str, zval *zv, int priority, int indent)
+static ZEND_COLD void zend_ast_export_zval(smart_str *str, zval *zv, int priority, int indent)
 {
 	zend_long idx;
 	zend_string *key;
@@ -932,7 +1196,7 @@ static void zend_ast_export_zval(smart_str *str, zval *zv, int priority, int ind
 		case IS_DOUBLE:
 			key = zend_strpprintf(0, "%.*G", (int) EG(precision), Z_DVAL_P(zv));
 			smart_str_appendl(str, ZSTR_VAL(key), ZSTR_LEN(key));
-			zend_string_release(key);
+			zend_string_release_ex(key, 0);
 			break;
 		case IS_STRING:
 			smart_str_appendc(str, '\'');
@@ -960,9 +1224,6 @@ static void zend_ast_export_zval(smart_str *str, zval *zv, int priority, int ind
 			} ZEND_HASH_FOREACH_END();
 			smart_str_appendc(str, ']');
 			break;
-		case IS_CONSTANT:
-			smart_str_appendl(str, Z_STRVAL_P(zv), Z_STRLEN_P(zv));
-			break;
 		case IS_CONSTANT_AST:
 			zend_ast_export_ex(str, Z_ASTVAL_P(zv), priority, indent);
 			break;
@@ -970,7 +1231,7 @@ static void zend_ast_export_zval(smart_str *str, zval *zv, int priority, int ind
 	}
 }
 
-static void zend_ast_export_class_no_header(smart_str *str, zend_ast_decl *decl, int indent) {
+static ZEND_COLD void zend_ast_export_class_no_header(smart_str *str, zend_ast_decl *decl, int indent) {
 	if (decl->child[0]) {
 		smart_str_appends(str, " extends ");
 		zend_ast_export_ns_name(str, decl->child[0], 0, indent);
@@ -1027,7 +1288,7 @@ static void zend_ast_export_class_no_header(smart_str *str, zend_ast_decl *decl,
 		goto append_default_value; \
 	} while (0)
 
-static void zend_ast_export_ex(smart_str *str, zend_ast *ast, int priority, int indent)
+static ZEND_COLD void zend_ast_export_ex(smart_str *str, zend_ast *ast, int priority, int indent)
 {
 	zend_ast_decl *decl;
 	int p, pl, pr;
@@ -1042,6 +1303,14 @@ tail_call:
 		case ZEND_AST_ZVAL:
 			zend_ast_export_zval(str, zend_ast_get_zval(ast), priority, indent);
 			break;
+		case ZEND_AST_CONSTANT: {
+			zend_string *name = zend_ast_get_constant_name(ast);
+			smart_str_appendl(str, ZSTR_VAL(name), ZSTR_LEN(name));
+			break;
+		}
+		case ZEND_AST_CONSTANT_CLASS:
+			smart_str_appendl(str, "__CLASS__", sizeof("__CLASS__")-1);
+			break;
 		case ZEND_AST_ZNODE:
 			/* This AST kind is only used for temporary nodes during compilation */
 			ZEND_ASSERT(0);
@@ -1194,7 +1463,7 @@ simple_list:
 			}
 			break;
 		case ZEND_AST_TYPE:
-			switch (ast->attr) {
+			switch (ast->attr & ~ZEND_TYPE_NULLABLE) {
 				case IS_ARRAY:    APPEND_STR("array");
 				case IS_CALLABLE: APPEND_STR("callable");
 				EMPTY_SWITCH_DEFAULT_CASE();
@@ -1724,7 +1993,7 @@ append_default_value:
 	return;
 }
 
-ZEND_API zend_string *zend_ast_export(const char *prefix, zend_ast *ast, const char *suffix)
+ZEND_API ZEND_COLD zend_string *zend_ast_export(const char *prefix, zend_ast *ast, const char *suffix)
 {
 	smart_str str = {0};
 
